# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
# Copyright (C) 2024 California Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations datastreams, transformers, writers and readers."""

from flask import current_app

from ...datastreams.errors import TransformerError, WriterError
from ...datastreams.transformers import BaseTransformer
from ...datastreams.writers import ServiceWriter
from ..common.ror.datastreams import RORTransformer


class AffiliationsServiceWriter(ServiceWriter):
    """Affiliations service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "affiliations")
        super().__init__(service_or_name=service_or_name, *args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


class AffiliationsRORTransformer(RORTransformer):
    """Affiliations ROR Transformer."""

    def __init__(
        self, *args, vocab_schemes=None, funder_fundref_doi_prefix=None, **kwargs
    ):
        """Constructor."""
        if vocab_schemes is None:
            vocab_schemes = current_app.config.get("VOCABULARIES_AFFILIATION_SCHEMES")
        super().__init__(
            *args,
            vocab_schemes=vocab_schemes,
            funder_fundref_doi_prefix=funder_fundref_doi_prefix,
            **kwargs,
        )


class OpenAIREOrganizationTransformer(BaseTransformer):
    """OpenAIRE Organization Transformer."""

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry

        if "id" not in record:
            raise TransformerError([f"No id for: {record}"])

        if not record["id"].startswith("openorgs____::"):
            raise TransformerError([f"Not valid OpenAIRE OpenOrgs id for: {record}"])

        if "pid" not in record:
            raise TransformerError([f"No alternative identifiers for: {record}"])

        organization = {}

        for pid in record["pid"]:
            if pid["scheme"] == "ROR":
                organization["id"] = pid["value"].removeprefix("https://ror.org/")
            elif pid["scheme"] == "PIC":
                organization["identifiers"] = [
                    {
                        "scheme": "pic",
                        "identifier": pid["value"],
                    }
                ]

        stream_entry.entry = organization
        return stream_entry


class OpenAIREAffiliationsServiceWriter(ServiceWriter):
    """OpenAIRE Affiliations service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "affiliations")
        # Here we only update and we do not insert, since OpenAIRE data is used to augment existing affiliations
        # (with PIC identifiers) and is not used to create new affiliations.
        super().__init__(
            service_or_name=service_or_name, insert=False, update=True, *args, **kwargs
        )

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]

    def write(self, stream_entry, *args, **kwargs):
        """Writes the input entry using a given service."""
        entry = stream_entry.entry

        return super().write(stream_entry, *args, **kwargs)

    def write_many(self, stream_entries, *args, **kwargs):
        """Writes the input entries using a given service."""
        return super().write_many(stream_entries, *args, **kwargs)


VOCABULARIES_DATASTREAM_READERS = {}
"""Affiliations datastream readers."""

VOCABULARIES_DATASTREAM_WRITERS = {
    "affiliations-service": AffiliationsServiceWriter,
    "openaire-affiliations-service": OpenAIREAffiliationsServiceWriter,
}
"""Affiliations datastream writers."""

VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "ror-affiliations": AffiliationsRORTransformer,
    "openaire-organization": OpenAIREOrganizationTransformer,
}
"""Affiliations datastream transformers."""


DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "zip",
            "args": {
                "regex": "_schema_v2\\.json$",
            },
        },
        {"type": "json"},
    ],
    "transformers": [
        {
            "type": "ror-affiliations",
        },
    ],
    "writers": [
        {
            "type": "async",
            "args": {
                "writer": {
                    "type": "affiliations-service",
                }
            },
        }
    ],
}
"""Data Stream configuration.

An origin is required for the reader.
"""

DATASTREAM_CONFIG_OPENAIRE = {
    "readers": [
        {"type": "openaire-http", "args": {"tar_href": "/organization.tar"}},
        {
            "type": "tar",
            "args": {
                "regex": "\\.json.gz$",
                "mode": "r",
            },
        },
        {"type": "gzip"},
        {"type": "jsonl"},
    ],
    "transformers": [
        {
            "type": "openaire-organization",
        },
    ],
    "writers": [
        {
            "type": "async",
            "args": {
                "writer": {
                    "type": "openaire-affiliations-service",
                }
            },
        }
    ],
}
"""Alternative Data Stream configuration for OpenAIRE Affiliations."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-RDM-Records is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.


"""Custom fields."""

from invenio_i18n import lazy_gettext as _

from ...contrib.subjects.api import Subject
from ...contrib.subjects.schema import SubjectRelationSchema
from .vocabulary import VocabularyCF


class SubjectCF(VocabularyCF):
    """Custom field for subjects."""

    field_keys = ["id", "subject"]

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(
            vocabulary_id="subjects",
            schema=SubjectRelationSchema,
            ui_schema=SubjectRelationSchema,
            **kwargs,
        )
        self.pid_field = Subject.pid

    @property
    def mapping(self):
        """Return the mapping."""
        _mapping = {
            "type": "object",
            "properties": {
                "@v": {"type": "keyword"},
                "id": {"type": "keyword"},
                "subject": {"type": "keyword"},
            },
        }

        return _mapping


SUBJECT_FIELDS_UI = [
    {
        "section": _("Subjects"),
        "fields": [
            dict(
                field="subjects",
                ui_widget="SubjectAutocompleteDropdown",
                isGenericVocabulary=False,
                props=dict(
                    label="Keywords and subjects",
                    icon="tag",
                    description="The subjects related to the community",
                    placeholder="Search for a subject by name e.g. Psychology ...",
                    autocompleteFrom="api/subjects",
                    noQueryMessage="Search for subjects...",
                    autocompleteFromAcceptHeader="application/vnd.inveniordm.v1+json",
                    required=False,
                    multiple=True,
                    clearable=True,
                    allowAdditions=False,
                ),
                template="invenio_vocabularies/subjects.html",
            )
        ],
    }
]


SUBJECT_FIELDS = {
    SubjectCF(
        name="subjects",
        multiple=True,
        dump_options=False,
    )
}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest
from invenio_db import db
from invenio_records_resources.proxies import current_service_registry

from invenio_vocabularies.contrib.funders.api import Funder


@pytest.fixture(scope="function")
def funder_full_data():
    """Full funder data."""
    return {
        "id": "01ggx4157",
        "identifiers": [
            {"scheme": "ror", "identifier": "01ggx4157"},
            {
                "identifier": "0000 0001 2156 142X",
                "scheme": "isni",
            },
            {
                "identifier": "grid.9132.9",
                "scheme": "grid",
            },
            {"scheme": "doi", "identifier": "10.13039/100012470"},
        ],
        "acronym": "CERN",
        "name": "European Organization for Nuclear Research",
        "title": {
            "en": "European Organization for Nuclear Research",
            "fr": "Organisation européenne pour la recherche nucléaire",
        },
        "country": "CH",
        "country_name": "Switzerland",
        "location_name": "Geneva",
        "status": "active",
        "types": ["facility", "funder"],
    }


@pytest.fixture(scope="module")
def service():
    """Funders service object."""
    return current_service_registry.get("funders")


@pytest.fixture()
def indexer(service):
    """Indexer instance with correct Record class."""
    return service.indexer


@pytest.fixture(scope="function")
def example_funder(identity, service, funder_full_data, indexer):
    """Creates and hard deletes a funder."""
    funder = service.create(identity, funder_full_data)
    Funder.index.refresh()  # Refresh the index
    yield funder
    funder._record.delete(force=True)
    indexer.delete(funder._record, refresh=True)
    db.session.commit()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""ROR-related Datastreams Readers/Writers/Transformers tests."""

import io
from unittest.mock import patch

import pytest

from invenio_vocabularies.contrib.common.ror.datastreams import (
    RORHTTPReader,
    RORTransformer,
)
from invenio_vocabularies.datastreams.errors import ReaderError

API_JSON_RESPONSE_CONTENT = {
    "linkset": [
        {
            "anchor": "https://zenodo.org/records/11186879",
            "describedby": [
                {
                    "href": "https://zenodo.org/api/records/11186879",
                    "type": "application/ld+json",
                },
            ],
            "item": [
                {
                    "href": "https://example.com/records/11186879/files/v1.46.1-2024-05-13-ror-data.zip",
                    "type": "application/zip",
                }
            ],
            "type": [{"href": "https://schema.org/Dataset"}],
        },
        {
            "anchor": "https://example.com/records/11186879/files/v1.46.1-2024-05-13-ror-data.zip",
            "collection": [
                {"href": "https://zenodo.org/records/11186879", "type": "text/html"}
            ],
        },
        {
            "anchor": "https://zenodo.org/api/records/11186879",
            "describes": [
                {"href": "https://zenodo.org/records/11186879", "type": "text/html"}
            ],
        },
    ]
}

API_JSON_RESPONSE_CONTENT_WITHOUT_JSON_LD = {
    "linkset": [
        {
            "anchor": "https://example.com/records/11186879",
            "item": [
                {
                    "href": "https://example.com/records/11186879/files/v1.46.1-2024-05-13-ror-data.zip",
                    "type": "application/zip",
                }
            ],
            "type": "application/zip",
        },
        {
            "anchor": "https://example.com/api/records/11186879",
            "describes": [
                {"href": "https://example.com/records/11186879", "type": "text/html"}
            ],
            "type": "application/dcat+xml",
        },
    ]
}

API_JSON_RESPONSE_CONTENT_WRONG_NUMBER_ZIP_ITEMS_ERROR = {
    "linkset": [
        {
            "anchor": "https://example.com/records/11186879",
            "item": [
                {
                    "href": "https://example.com/records/11186879/files/v1.46.1-2024-05-13-ror-data.zip",
                    "type": "application/zip",
                },
                {
                    "href": "https://example.com/another/file.zip",
                    "type": "application/zip",
                },
            ],
        },
        {
            "anchor": "https://example.com/api/records/11186879",
            "describes": [
                {"href": "https://example.com/records/11186879", "type": "text/html"}
            ],
            "type": "application/dcat+xml",
        },
    ]
}


API_JSON_RESPONSE_CONTENT_LD_JSON = {
    "name": "ROR Data",
    "dateCreated": "2024-07-11T22:29:25.727626+00:00",
    "datePublished": "2024-07-11T22:29:25.727626+00:00",
}

DOWNLOAD_FILE_BYTES_CONTENT = b"The content of the file"

DOI_LINKSET_INFO = {
    "linkset": {
        "url": "https://zenodo.org/api/records/12729557",
        "rel": "linkset",
        "type": "application/linkset+json",
    }
}


@pytest.fixture(scope="module")
def expected_from_ror_json():
    return {
        "id": "05dxps055",
        "name": "California Institute of Technology",
        "title": {
            "en": "California Institute of Technology",
            "es": "Instituto de Tecnología de California",
        },
        "acronym": "CIT",
        "aliases": ["Caltech"],
        "country": "US",
        "country_name": "United States",
        "location_name": "Pasadena",
        "status": "active",
        "identifiers": [
            {"scheme": "ror", "identifier": "05dxps055"},
        ],
        "types": ["education", "funder"],
    }


class MockResponse:
    content = DOWNLOAD_FILE_BYTES_CONTENT
    links = DOI_LINKSET_INFO

    def __init__(self, api_json_response_content, remove_links=False):
        self.api_json_response_content = api_json_response_content
        self.links = DOI_LINKSET_INFO
        if remove_links:
            self.links = {}

    def json(self, **kwargs):
        return self.api_json_response_content

    def raise_for_status(self):
        pass


def side_effect(url, headers=None, allow_redirects=False):
    if not headers:
        return MockResponse({})
    if headers["Accept"] == "application/ld+json":
        return MockResponse(API_JSON_RESPONSE_CONTENT_LD_JSON)
    elif headers["Accept"] == "application/linkset+json":
        return MockResponse(API_JSON_RESPONSE_CONTENT)


@patch("requests.get", side_effect=side_effect)
def test_ror_http_reader(_):
    reader = RORHTTPReader()
    results = []
    for entry in reader.read():
        results.append(entry)

    assert len(results) == 1
    assert isinstance(results[0], io.BytesIO)
    assert results[0].read() == DOWNLOAD_FILE_BYTES_CONTENT


@patch("requests.get", side_effect=side_effect)
def test_ror_http_reader_since_before_publish(_):
    reader = RORHTTPReader(since="2024-07-10")
    results = []
    for entry in reader.read():
        results.append(entry)

    assert len(results) == 1


@patch("requests.get", side_effect=side_effect)
def test_ror_http_reader_since_after_publish(_):
    reader = RORHTTPReader(since="2024-07-12")
    results = []
    for entry in reader.read():
        results.append(entry)

    assert len(results) == 0


@patch(
    "requests.get",
    side_effect=lambda url, headers=None, allow_redirects=False: MockResponse(
        API_JSON_RESPONSE_CONTENT_WRONG_NUMBER_ZIP_ITEMS_ERROR
    ),
)
def test_ror_http_reader_wrong_number_zip_items_error(_):
    reader = RORHTTPReader()
    with pytest.raises(ReaderError):
        next(reader.read())


@patch(
    "requests.get",
    side_effect=lambda url, headers=None, allow_redirects=False: MockResponse(
        API_JSON_RESPONSE_CONTENT_WITHOUT_JSON_LD
    ),
)
def test_ror_http_reader_no_json_ld(_):
    reader = RORHTTPReader(since="2024-07-12")
    with pytest.raises(ReaderError):
        next(reader.read())


def test_ror_http_reader_item_not_implemented():
    reader = RORHTTPReader()
    with pytest.raises(NotImplementedError):
        next(reader.read("A fake item"))


def test_ror_http_reader_iter_not_implemented():
    reader = RORHTTPReader()
    with pytest.raises(NotImplementedError):
        reader._iter("A fake file pointer")


def test_ror_transformer(app, dict_ror_entry, expected_from_ror_json):
    transformer = RORTransformer()
    assert expected_from_ror_json == transformer.apply(dict_ror_entry).entry



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Vocabulariess is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test award schema."""

import pytest
from marshmallow import ValidationError

from invenio_vocabularies.contrib.awards.schema import (
    AwardRelationSchema,
    AwardSchema,
    FundingRelationSchema,
)


#
# AwardRelationSchema
#
def test_valid_full(app, award_full_data):
    loaded = AwardSchema().load(award_full_data)
    award_full_data["pid"] = award_full_data.pop("id")
    assert award_full_data == loaded


def test_valid_minimal():
    valid_minimal = {"id": "755021", "number": "755021"}
    assert {
        "pid": "755021",
        "number": "755021",
    } == AwardSchema().load(valid_minimal)


def test_invalid_no_number(app):
    invalid_no_number = {
        "identifiers": [{"identifier": "10.5281/zenodo.9999999", "scheme": "doi"}]
    }
    with pytest.raises(ValidationError):
        data = AwardSchema().load(invalid_no_number)


def test_invalid_empty_number(app):
    invalid_empty_number = {"id": "755021", "number": ""}

    with pytest.raises(ValidationError):
        data = AwardSchema().load(invalid_empty_number)


def test_invalid_empty_pid(app):
    invalid_empty_pid = {"id": "", "number": "755021"}

    with pytest.raises(ValidationError):
        AwardSchema().load(invalid_empty_pid)


def test_award_funder_name(app):
    with_funder_name = {
        "id": "755021",
        "number": "755021",
        "funder": {
            "name": "custom funder",
        },
    }
    assert {
        "pid": "755021",
        "number": "755021",
        "funder": {
            "name": "custom funder",
        },
    } == AwardSchema().load(with_funder_name)


def test_award_funder_id(app):
    with_funder_id = {
        "id": "755021",
        "number": "755021",
        "funder": {
            "id": "01ggx4157",
        },
    }
    assert {
        "pid": "755021",
        "number": "755021",
        "funder": {
            "id": "01ggx4157",
        },
    } == AwardSchema().load(with_funder_id)


#
# AwardRelationSchema
#


def test_valid_id():
    valid_id = {
        "id": "test",
    }
    assert valid_id == AwardRelationSchema().load(valid_id)


def test_valid_number_title():
    valid_data = {"number": "123456", "title": {"en": "Test title."}}
    assert valid_data == AwardRelationSchema().load(valid_data)


def test_invalid_empty():
    invalid_data = {}
    with pytest.raises(ValidationError):
        data = AwardRelationSchema().load(invalid_data)


def test_invalid_number_type():
    invalid_data = {"number": 123}
    with pytest.raises(ValidationError):
        data = AwardRelationSchema().load(invalid_data)


#
# FundingRelationSchema
#


AWARD = {"title": {"en": "Some award"}, "number": "755021"}

FUNDER = {
    "name": "Someone",
}


def test_valid_award_funding():
    valid_funding = {"award": AWARD}
    assert valid_funding == FundingRelationSchema().load(valid_funding)

    # Test a valid award with different representation
    valid_funding = {"award": {"id": "test-award-id"}}
    assert valid_funding == FundingRelationSchema().load(valid_funding)


def test_invalid_award_funding():
    invalid_funding = {"award": {"identifiers": [AWARD.get("identifiers")]}}
    with pytest.raises(ValidationError):
        data = FundingRelationSchema().load(invalid_funding)


def test_valid_funder_funding():
    valid_funding = {"funder": FUNDER}
    assert valid_funding == FundingRelationSchema().load(valid_funding)

    # Test a valid funder with different representation
    valid_funding = {"funder": {"id": "test-funder-id"}}
    assert valid_funding == FundingRelationSchema().load(valid_funding)


def test_invalid_funder_funding():
    invalid_funding = {"funder": {"identifiers": [AWARD.get("identifiers")]}}
    with pytest.raises(ValidationError):
        data = FundingRelationSchema().load(invalid_funding)


def test_valid_award_funder_funding():
    valid_funding = {"funder": FUNDER, "award": AWARD}
    assert valid_funding == FundingRelationSchema().load(valid_funding)


def test_invalid_empty_funding():
    invalid_funding = {}
    with pytest.raises(ValidationError):
        data = FundingRelationSchema().load(invalid_funding)
